from __future__ import absolute_import, division, print_function

__all__ = [
    "__title__", "__summary__", "__uri__", "__version__", "__author__",
    "__email__", "__license__", "__copyright__",
]

__title__ = "phe_cuda"
__summary__ = "Acceleration for Partially Homomorphic Encryption library by Cuda "
__uri__ = "https://github.com/gxx777/phe_cuda"

# We use semantic versioning - semver.org
__version__ = "1.0.0"

__author__ = "Xiangxin Guo,WenTian Zhu,Yonglin Wu"
__email__ = "xiangxinguo@mail.ustc.edu.cn"
__license__ = "GPLv3"
__copyright__ = "Copyright 2022 {0}".format(__author__)
